@@ -19,9 +19,11 @@ import sys
 import traceback
 import urlparse
 import zipfile
+import xml.dom.minidom
+import mimetypes
 
 PROG = os.path.basename(os.path.splitext(__file__)[0])
-VERSION = '8.6.1'
+VERSION = '8.6.7'
 
 # AsciiDoc global configuration file directory.
 # NOTE: CONF_DIR is "fixed up" by Makefile -- don't rename or change syntax.
@@ -51,6 +53,7 @@ ASCIIDOC_OPTS = ''
 DBLATEX_OPTS = ''
 FOP_OPTS = ''
 XSLTPROC_OPTS = ''
+BACKEND_OPTS = ''
 
 ######################################################################
 # End of configuration file parameters.
@@ -141,6 +144,20 @@ def find_executable(file_name):
         result = _find_executable(file_name)
     return result
 
+def write_file(filename, data, mode='w'):
+    f = open(filename, mode)
+    try:
+        f.write(data)
+    finally:
+        f.close()
+
+def read_file(filename, mode='r'):
+    f = open(filename, mode)
+    try:
+        return f.read()
+    finally:
+        f.close()
+
 def shell_cd(path):
     verbose('chdir %s' % path)
     if not OPTIONS.dry_run:
@@ -174,7 +191,8 @@ def shell_rmtree(path):
 
 def shell(cmd, raise_error=True):
     '''
-    Execute command cmd in shell and return resulting subprocess.Popen object.
+    Execute command cmd in shell and return tuple
+    (stdoutdata, stderrdata, returncode).
     If raise_error is True then a non-zero return terminates the application.
     '''
     if os.name == 'nt':
@@ -188,25 +206,26 @@ def shell(cmd, raise_error=True):
             mo = re.match(r'^\s*(?P<arg0>[^ ]+)', cmd)
         if mo.group('arg0').endswith('.py'):
             cmd = 'python ' + cmd
-    # Remove redundant quoting -- this is not just costmetic, quoting seems to
-    # dramatically decrease the allowed command length in Windows XP.
-    cmd = re.sub(r'"([^ ]+?)"', r'\1', cmd)
+        # Remove redundant quoting -- this is not just cosmetic,
+        # quoting seems to dramatically decrease the allowed command
+        # length in Windows XP.
+        cmd = re.sub(r'"([^ ]+?)"', r'\1', cmd)
     verbose('executing: %s' % cmd)
     if OPTIONS.dry_run:
         return
-    if OPTIONS.verbose:
-        stdout = stderr = None
-    else:
-        stdout = stderr = subprocess.PIPE
+    stdout = stderr = subprocess.PIPE
     try:
         popen = subprocess.Popen(cmd, stdout=stdout, stderr=stderr,
                 shell=True, env=ENV)
     except OSError, e:
         die('failed: %s: %s' % (cmd, e))
-    popen.wait()
+    stdoutdata, stderrdata = popen.communicate()
+    if OPTIONS.verbose:
+        print stdoutdata
+        print stderrdata
     if popen.returncode != 0 and raise_error:
         die('%s returned non-zero exit status %d' % (cmd, popen.returncode))
-    return popen
+    return (stdoutdata, stderrdata, popen.returncode)
 
 def find_resources(files, tagname, attrname, filter=None):
     '''
@@ -232,12 +251,20 @@ def find_resources(files, tagname, attrname, filter=None):
     if isinstance(files, str):
         files = [files]
     result = []
-    for f in files:
-        verbose('finding resources in: %s' % f)
+    for filename in files:
+        verbose('finding resources in: %s' % filename)
         if OPTIONS.dry_run:
             continue
         parser = FindResources()
-        parser.feed(open(f).read())
+        # HTMLParser has problems with non-ASCII strings.
+        # See http://bugs.python.org/issue3932
+        contents = read_file(filename)
+        mo = re.search(r'\A<\?xml.* encoding="(.*?)"', contents)
+        if mo:
+            encoding = mo.group(1)
+            parser.feed(contents.decode(encoding))
+        else:
+            parser.feed(contents)
         parser.close()
     result = list(set(result))   # Drop duplicate values.
     result.sort()
@@ -248,12 +275,12 @@ def copy_files(files, src_dir, dst_dir):
     '''
     Copy list of relative file names from src_dir to dst_dir.
     '''
-    for f in files:
-        f = os.path.normpath(f)
-        if os.path.isabs(f):
+    for filename in files:
+        filename = os.path.normpath(filename)
+        if os.path.isabs(filename):
             continue
-        src = os.path.join(src_dir, f)
-        dst = os.path.join(dst_dir, f)
+        src = os.path.join(src_dir, filename)
+        dst = os.path.join(dst_dir, filename)
         if not os.path.exists(dst):
             if not os.path.isfile(src):
                 warning('missing file: %s' % src)
@@ -312,10 +339,14 @@ def get_source_options(asciidoc_file):
     result = []
     if os.path.isfile(asciidoc_file):
         options = ''
-        for line in open(asciidoc_file):
-            mo = re.search(r'^//\s*a2x:', line)
-            if mo:
-                options += ' ' + line[mo.end():].strip()
+        f = open(asciidoc_file)
+        try:
+            for line in f:
+                mo = re.search(r'^//\s*a2x:', line)
+                if mo:
+                    options += ' ' + line[mo.end():].strip()
+        finally:
+            f.close()
         parse_options()
     return result
 
@@ -339,8 +370,12 @@ class A2X(AttrDict):
         self.dblatex_opts  += ' ' + DBLATEX_OPTS
         self.fop_opts      += ' ' + FOP_OPTS
         self.xsltproc_opts += ' ' + XSLTPROC_OPTS
+        self.backend_opts  += ' ' + BACKEND_OPTS
         # Execute to_* functions.
-        self.__getattribute__('to_'+self.format)()
+        if self.backend:
+            self.to_backend()
+        else:
+            self.__getattribute__('to_'+self.format)()
         if not (self.keep_artifacts or self.format == 'docbook' or self.skip_asciidoc):
             shell_rm(self.dst_path('.xml'))
 
@@ -369,21 +404,38 @@ class A2X(AttrDict):
         home_dir = os.environ.get('HOME')
         if home_dir is not None:
             conf_files.append(os.path.join(home_dir, '.asciidoc', CONF_FILE))
-        # From --conf-file option.
-        if self.conf_file is not None:
-            if not os.path.isfile(self.conf_file):
-                die('missing configuration file: %s' % self.conf_file)
-            conf_files.append(self.conf_file)
-        # Load ordered files.
-        for f in conf_files:
-            if os.path.isfile(f):
-                verbose('loading conf file: %s' % f)
-                execfile(f, globals())
         # If asciidoc is not local to a2x then search the PATH.
         if not self.asciidoc:
             self.asciidoc = find_executable(ASCIIDOC)
             if not self.asciidoc:
                 die('unable to find asciidoc: %s' % ASCIIDOC)
+        # From backend plugin directory.
+        if self.backend is not None:
+            stdout = shell(self.asciidoc + ' --backend list')[0]
+            backends = [(i, os.path.split(i)[1]) for i in stdout.splitlines()]
+            backend_dir = [i[0] for i in backends if i[1] == self.backend]
+            if len(backend_dir) == 0:
+                die('missing %s backend' % self.backend)
+            if len(backend_dir) > 1:
+                die('more than one %s backend' % self.backend)
+            verbose('found %s backend directory: %s' %
+                    (self.backend, backend_dir[0]))
+            conf_files.append(os.path.join(backend_dir[0], 'a2x-backend.py'))
+        # From --conf-file option.
+        if self.conf_file is not None:
+            if not os.path.isfile(self.conf_file):
+                die('missing configuration file: %s' % self.conf_file)
+            conf_files.append(self.conf_file)
+        # From --xsl-file option.
+        if self.xsl_file is not None:
+            if not os.path.isfile(self.xsl_file):
+                die('missing XSL file: %s' % self.xsl_file)
+            self.xsl_file = os.path.abspath(self.xsl_file)
+        # Load ordered files.
+        for f in conf_files:
+            if os.path.isfile(f):
+                verbose('loading configuration file: %s' % f)
+                execfile(f, globals())
 
     def process_options(self):
         '''
@@ -403,26 +455,41 @@ class A2X(AttrDict):
         if self.resource_manifest:
             if not os.path.isfile(self.resource_manifest):
                 die('missing --resource-manifest: %s' % self.resource_manifest)
-            for r in open(self.resource_manifest):
-                self.resources.append(r.strip())
+            f = open(self.resource_manifest)
+            try:
+                for r in f:
+                    self.resources.append(r.strip())
+            finally:
+                f.close()
         for r in self.resources:
-            if os.path.isfile(r):
-                self.resource_files.append(r)
+            r = os.path.expanduser(r)
+            r = os.path.expandvars(r)
+            if r.endswith('/') or r.endswith('\\'):
+                if  os.path.isdir(r):
+                    self.resource_dirs.append(r)
+                else:
+                    die('missing resource directory: %s' % r)
             elif os.path.isdir(r):
                 self.resource_dirs.append(r)
+            elif r.startswith('.') and '=' in r:
+                ext, mimetype = r.split('=')
+                mimetypes.add_type(mimetype, ext)
             else:
-                die('missing resource: %s' % r)
-        # Lastly search among images and stylesheets distributed with asciidoc.
+                self.resource_files.append(r)
         for p in (os.path.dirname(self.asciidoc), CONF_DIR):
             for d in ('images','stylesheets'):
-                self.resource_dirs.append(os.path.join(p,d))
+                d = os.path.join(p,d)
+                if os.path.isdir(d):
+                    self.resource_dirs.append(d)
+        verbose('resource files: %s' % self.resource_files)
+        verbose('resource directories: %s' % self.resource_dirs)
         if not self.doctype and self.format == 'manpage':
             self.doctype = 'manpage'
         if self.doctype:
             self.asciidoc_opts += ' --doctype %s' % self.doctype
         for attr in self.attributes:
             self.asciidoc_opts += ' --attribute "%s"' % attr
-        self.xsltproc_opts += ' --nonet'
+#        self.xsltproc_opts += ' --nonet'
         if self.verbose:
             self.asciidoc_opts += ' --verbose'
             self.dblatex_opts += ' -V'
@@ -447,9 +514,19 @@ class A2X(AttrDict):
                 'admon.graphics 0',
             ]
         if self.stylesheet:
-            params.append('html.stylesheet "%s"' % self.stylesheet)
-        params = ['--stringparam %s' % o for o in params]
-        self.xsltproc_opts += ' ' + ' '.join(params)
+            params += ['html.stylesheet "%s"' % self.stylesheet]
+        if self.format == 'htmlhelp':
+            params += ['htmlhelp.chm "%s"' % self.basename('.chm'),
+                       'htmlhelp.hhp "%s"' % self.basename('.hhp'),
+                       'htmlhelp.hhk "%s"' % self.basename('.hhk'),
+                       'htmlhelp.hhc "%s"' % self.basename('.hhc')]
+        if self.doctype == 'book':
+            params += ['toc.section.depth 1']
+            # Books are chunked at chapter level.
+            params += ['chunk.section.depth 0']
+        for o in params:
+            if o.split()[0]+' ' not in self.xsltproc_opts:
+                self.xsltproc_opts += ' --stringparam ' + o
         if self.fop_opts:
             self.fop = True
         if os.path.splitext(self.asciidoc_file)[1].lower() == '.xml':
@@ -484,11 +561,15 @@ class A2X(AttrDict):
                 die('missing configuration file: %s' % f)
         return os.path.normpath(f)
 
-    def xsl_file(self, file_name=None):
+    def xsl_stylesheet(self, file_name=None):
         '''
         Return full path name of file in asciidoc docbook-xsl configuration
         directory.
+        If an XSL file was specified with the --xsl-file option then it is
+        returned.
         '''
+        if self.xsl_file is not None:
+            return self.xsl_file
         if not file_name:
             file_name = self.format + '.xsl'
         return self.asciidoc_conf_file(os.path.join('docbook-xsl', file_name))
@@ -497,12 +578,10 @@ class A2X(AttrDict):
         '''
         Search html_files for images and CSS resource URIs (html_files can be a
         list of file names or a single file name).
-        If the URIs are relative then copy them from the src_dir to the
-        dst_dir.
+        Copy them from the src_dir to the dst_dir.
         If not found in src_dir then recursively search all specified
-        resources for the file name.
-        Optional additional resources URIs can be passed in the resources list.
-        Does not copy absolute URIs.
+        resource directories.
+        Optional additional resources files can be passed in the resources list.
         '''
         resources = resources[:]
         resources += find_resources(html_files, 'link', 'href',
@@ -512,29 +591,46 @@ class A2X(AttrDict):
         resources = list(set(resources))    # Drop duplicates.
         resources.sort()
         for f in resources:
-            f = os.path.normpath(f)
-            if os.path.isabs(f):
-                if not os.path.isfile(f):
-                    warning('missing resource: %s' % f)
-                continue
-            src = os.path.join(src_dir, f)
-            dst = os.path.join(dst_dir, f)
+            if '=' in f:
+                src, dst = f.split('=')
+                if not dst:
+                    dst = src
+            else:
+                src = dst = f
+            src = os.path.normpath(src)
+            dst = os.path.normpath(dst)
+            if os.path.isabs(dst):
+                die('absolute resource file name: %s' % dst)
+            if dst.startswith(os.pardir):
+                die('resource file outside destination directory: %s' % dst)
+            src = os.path.join(src_dir, src)
+            dst = os.path.join(dst_dir, dst)
             if not os.path.isfile(src):
                 for d in self.resource_dirs:
-                    src = find_files(d, os.path.basename(f))
-                    if src:
-                        src = src[0]
+                    d = os.path.join(src_dir, d)
+                    found = find_files(d, os.path.basename(src))
+                    if found:
+                        src = found[0]
                         break
                 else:
                     if not os.path.isfile(dst):
-                        warning('missing resource: %s' % f)
-                    continue    # Continues outer for loop.
-            # Arrive here if relative resource file has been found.
+                        die('missing resource: %s' % src)
+                    continue
+            # Arrive here if resource file has been found.
             if os.path.normpath(src) != os.path.normpath(dst):
                 dstdir = os.path.dirname(dst)
                 shell_makedirs(dstdir)
                 shell_copy(src, dst)
 
+    def to_backend(self):
+        '''
+        Convert AsciiDoc source file to a backend output file using the global
+        'to_<backend name>' function (loaded from backend plugin a2x-backend.py
+        file).
+        Executes the global function in an A2X class instance context.
+        '''
+        eval('to_%s(self)' % self.backend)
+
     def to_docbook(self):
         '''
         Use asciidoc to convert asciidoc_file to DocBook.
@@ -545,8 +641,8 @@ class A2X(AttrDict):
             if not os.path.isfile(docbook_file):
                 die('missing docbook file: %s' % docbook_file)
             return
-        shell('"%s" --backend docbook %s --out-file "%s" "%s"' %
-             (self.asciidoc, self.asciidoc_opts, docbook_file, self.asciidoc_file))
+        shell('"%s" --backend docbook -a "a2x-format=%s" %s --out-file "%s" "%s"' %
+             (self.asciidoc, self.format, self.asciidoc_opts, docbook_file, self.asciidoc_file))
         if not self.no_xmllint and XMLLINT:
             shell('"%s" --nonet --noout --valid "%s"' % (XMLLINT, docbook_file))
 
@@ -555,7 +651,7 @@ class A2X(AttrDict):
         docbook_file = self.dst_path('.xml')
         xhtml_file = self.dst_path('.html')
         opts = '%s --output "%s"' % (self.xsltproc_opts, xhtml_file)
-        exec_xsltproc(self.xsl_file(), docbook_file, self.destination_dir, opts)
+        exec_xsltproc(self.xsl_stylesheet(), docbook_file, self.destination_dir, opts)
         src_dir = os.path.dirname(self.asciidoc_file)
         self.copy_resources(xhtml_file, src_dir, self.destination_dir)
 
@@ -563,7 +659,7 @@ class A2X(AttrDict):
         self.to_docbook()
         docbook_file = self.dst_path('.xml')
         opts = self.xsltproc_opts
-        exec_xsltproc(self.xsl_file(), docbook_file, self.destination_dir, opts)
+        exec_xsltproc(self.xsl_stylesheet(), docbook_file, self.destination_dir, opts)
 
     def to_pdf(self):
         if self.fop:
@@ -574,7 +670,7 @@ class A2X(AttrDict):
     def exec_fop(self):
         self.to_docbook()
         docbook_file = self.dst_path('.xml')
-        xsl = self.xsl_file('fo.xsl')
+        xsl = self.xsl_stylesheet('fo.xsl')
         fo = self.dst_path('.fo')
         pdf = self.dst_path('.pdf')
         opts = '%s --output "%s"' % (self.xsltproc_opts, fo)
@@ -607,15 +703,13 @@ class A2X(AttrDict):
         self.to_docbook()
         docbook_file = self.dst_path('.xml')
         opts = self.xsltproc_opts
-        xsl_file = self.xsl_file()
+        xsl_file = self.xsl_stylesheet()
         if self.format == 'chunked':
             dst_dir = self.dst_path('.chunked')
         elif self.format == 'htmlhelp':
             dst_dir = self.dst_path('.htmlhelp')
-            opts += ' --stringparam htmlhelp.chm "%s"' % self.basename('.chm')
-            opts += ' --stringparam htmlhelp.hhc "%s"' % self.basename('.hhc')
-            opts += ' --stringparam htmlhelp.hhp "%s"' % self.basename('.hhp')
-        opts += ' --stringparam base.dir "%s/"' % os.path.basename(dst_dir)
+        if not 'base.dir ' in opts:
+            opts += ' --stringparam base.dir "%s/"' % os.path.basename(dst_dir)
         # Create content.
         shell_rmtree(dst_dir)
         shell_makedirs(dst_dir)
@@ -624,9 +718,48 @@ class A2X(AttrDict):
         src_dir = os.path.dirname(self.asciidoc_file)
         self.copy_resources(html_files, src_dir, dst_dir)
 
+    def update_epub_manifest(self, opf_file):
+        '''
+        Scan the OEBPS directory for any files that have not been registered in
+        the OPF manifest then add them to the manifest.
+        '''
+        opf_dir = os.path.dirname(opf_file)
+        resource_files = []
+        for (p,dirs,files) in os.walk(os.path.dirname(opf_file)):
+            for f in files:
+                f = os.path.join(p,f)
+                if os.path.isfile(f):
+                    assert f.startswith(opf_dir)
+                    f = '.' + f[len(opf_dir):]
+                    f = os.path.normpath(f)
+                    if f not in ['content.opf']:
+                        resource_files.append(f)
+        opf = xml.dom.minidom.parseString(read_file(opf_file))
+        manifest_files = []
+        manifest = opf.getElementsByTagName('manifest')[0]
+        for el in manifest.getElementsByTagName('item'):
+            f = el.getAttribute('href')
+            f = os.path.normpath(f)
+            manifest_files.append(f)
+        count = 0
+        for f in resource_files:
+            if f not in manifest_files:
+                count += 1
+                verbose('adding to manifest: %s' % f)
+                item = opf.createElement('item')
+                item.setAttribute('href', f.replace(os.path.sep, '/'))
+                item.setAttribute('id', 'a2x-%d' % count)
+                mimetype = mimetypes.guess_type(f)[0]
+                if mimetype is None:
+                    die('unknown mimetype: %s' % f)
+                item.setAttribute('media-type', mimetype)
+                manifest.appendChild(item)
+        if count > 0:
+            write_file(opf_file, opf.toxml())
+
     def to_epub(self):
         self.to_docbook()
-        xsl_file = self.xsl_file()
+        xsl_file = self.xsl_stylesheet()
         docbook_file = self.dst_path('.xml')
         epub_file = self.dst_path('.epub')
         build_dir = epub_file + '.d'
@@ -634,16 +767,17 @@ class A2X(AttrDict):
         shell_makedirs(build_dir)
         # Create content.
         exec_xsltproc(xsl_file, docbook_file, build_dir, self.xsltproc_opts)
-        # Copy OPF file resources.
+        # Copy resources referenced in the OPF and resources referenced by the
+        # generated HTML (in theory DocBook XSL should ensure they are
+        # identical but this is not always the case).
         src_dir = os.path.dirname(self.asciidoc_file)
         dst_dir = os.path.join(build_dir, 'OEBPS')
-        # Get the resources from OPF instead of HTML content to get round this
-        # bug: https://sourceforge.net/tracker/?func=detail&aid=2854080&group_id=21935&atid=373747
-        opf = os.path.join(dst_dir, 'content.opf')
-        resources = find_resources(opf, 'item', 'href')
-#        html_files = find_files(dst_dir, '*.html')
-#        self.copy_resources(html_files, src_dir, dst_dir)
-        self.copy_resources([], src_dir, dst_dir, resources)
+        opf_file = os.path.join(dst_dir, 'content.opf')
+        opf_resources = find_resources(opf_file, 'item', 'href')
+        html_files = find_files(dst_dir, '*.html')
+        self.copy_resources(html_files, src_dir, dst_dir, opf_resources)
+        # Register any unregistered resources.
+        self.update_epub_manifest(opf_file)
         # Build epub archive.
         cwd = os.getcwd()
         shell_cd(build_dir)
@@ -653,7 +787,7 @@ class A2X(AttrDict):
                 try:
                     # Create and add uncompressed mimetype file.
                     verbose('archiving: mimetype')
-                    open('mimetype','w').write('application/epub+zip')
+                    write_file('mimetype', 'application/epub+zip')
                     zip.write('mimetype', compress_type=zipfile.ZIP_STORED)
                     # Compress all remaining files.
                     for (p,dirs,files) in os.walk('.'):
@@ -679,9 +813,9 @@ class A2X(AttrDict):
         text_file = self.dst_path('.text')
         html_file = self.dst_path('.text.html')
         if self.lynx:
-            shell('"%s" %s --conf-file "%s" -b html4 -o "%s" "%s"' %
+            shell('"%s" %s --conf-file "%s" -b html4 -a "a2x-format=%s" -o "%s" "%s"' %
                  (self.asciidoc, self.asciidoc_opts, self.asciidoc_conf_file('text.conf'),
-                  html_file, self.asciidoc_file))
+                  self.format, html_file, self.asciidoc_file))
             shell('"%s" -dump "%s" > "%s"' %
                  (LYNX, html_file, text_file))
         else:
@@ -689,7 +823,7 @@ class A2X(AttrDict):
             self.to_docbook()
             docbook_file = self.dst_path('.xml')
             opts = '%s --output "%s"' % (self.xsltproc_opts, html_file)
-            exec_xsltproc(self.xsl_file(), docbook_file,
+            exec_xsltproc(self.xsl_stylesheet(), docbook_file,
                     self.destination_dir, opts)
             shell('"%s" -cols 70 -dump -T text/html -no-graph "%s" > "%s"' %
                  (W3M, html_file, text_file))
@@ -711,7 +845,7 @@ if __name__ == '__main__':
         action='append', dest='attributes', default=[], metavar='ATTRIBUTE',
         help='set asciidoc attribute value')
     parser.add_option('--asciidoc-opts',
-        action='store', dest='asciidoc_opts', default='',
+        action='append', dest='asciidoc_opts', default=[],
         metavar='ASCIIDOC_OPTS', help='asciidoc options')
     #DEPRECATED
     parser.add_option('--copy',
@@ -727,6 +861,9 @@ if __name__ == '__main__':
         action='store', dest='doctype', metavar='DOCTYPE',
         choices=('article','manpage','book'),
         help='article, manpage, book')
+    parser.add_option('-b','--backend',
+        action='store', dest='backend', metavar='BACKEND',
+        help='name of backend plugin')
     parser.add_option('--epubcheck',
         action='store_true', dest='epubcheck', default=False,
         help='check EPUB output with epubcheck')
@@ -773,33 +910,44 @@ if __name__ == '__main__':
     parser.add_option('--stylesheet',
         action='store', dest='stylesheet', default=None,
         metavar='STYLESHEET',
-        help='target HTML CSS stylesheet file name')
+        help='HTML CSS stylesheet file name')
     #DEPRECATED
     parser.add_option('--safe',
         action='store_true', dest='safe', default=False,
         help='DEPRECATED: does nothing')
     parser.add_option('--dblatex-opts',
-        action='store', dest='dblatex_opts', default='',
+        action='append', dest='dblatex_opts', default=[],
         metavar='DBLATEX_OPTS', help='dblatex options')
+    parser.add_option('--backend-opts',
+        action='append', dest='backend_opts', default=[],
+        metavar='BACKEND_OPTS', help='backend plugin options')
     parser.add_option('--fop',
         action='store_true', dest='fop', default=False,
         help='use FOP to generate PDF files')
     parser.add_option('--fop-opts',
-        action='store', dest='fop_opts', default='',
+        action='append', dest='fop_opts', default=[],
         metavar='FOP_OPTS', help='options for FOP pdf generation')
     parser.add_option('--xsltproc-opts',
-        action='store', dest='xsltproc_opts', default='',
-        metavar='XSLTPROC_OPTS', help='options for FOP pdf generation')
+        action='append', dest='xsltproc_opts', default=[],
+        metavar='XSLTPROC_OPTS', help='xsltproc options for XSL stylesheets')
+    parser.add_option('--xsl-file',
+        action='store', dest='xsl_file', metavar='XSL_FILE',
+        help='custom XSL stylesheet')
     parser.add_option('-v', '--verbose',
         action='count', dest='verbose', default=0,
         help='increase verbosity')
     if len(sys.argv) == 1:
         parser.parse_args(['--help'])
     source_options = get_source_options(sys.argv[-1])
-    argv = source_options + sys.argv[1:] 
+    argv = source_options + sys.argv[1:]
     opts, args = parser.parse_args(argv)
     if len(args) != 1:
         parser.error('incorrect number of arguments')
+    opts.asciidoc_opts = ' '.join(opts.asciidoc_opts)
+    opts.dblatex_opts = ' '.join(opts.dblatex_opts)
+    opts.fop_opts = ' '.join(opts.fop_opts)
+    opts.xsltproc_opts = ' '.join(opts.xsltproc_opts)
+    opts.backend_opts = ' '.join(opts.backend_opts)
     opts = eval(str(opts))  # Convert optparse.Values to dict.
     a2x = A2X(opts)
     OPTIONS = a2x           # verbose and dry_run used by utility functions.
